(* A pretty-printer for F. *)

open PPrint
open ML

(* -------------------------------------------------------------------------- *)

(* Types. *)

let print_tyvar x =
  let rec aux x =
    if x < 26 then
      Char.escaped (Char.chr (x+97))
    else
      ("a_"^(string_of_int x))(*TEMPORAIRE*)
  in
  string "'"^^ string (aux x)

let print_tyvars tys =
  let rec aux acc = function
    | [] -> string "()"
    | ty :: [] -> parens (acc ^^ print_tyvar ty)
    | ty :: xs -> aux (acc ^^ print_tyvar ty ^^ string ", ") xs in
  aux empty tys

let rec print_args level args =
  match args with
  | [] ->  string ""
  | ty :: [] -> print_type_aux level ty
  | ty :: xs -> print_type_aux level ty ^^ (string " ,") ^^ print_args level xs

and print_annot tyopt =
  match tyopt with
  | None ->
    PPrint.empty
  | Some ty ->
    string " : " ^^
    print_type ty

and print_type_aux level ty =
  assert (level >= 0);
  match ty with
  | TyVar x ->
      print_tyvar x
  | TyCon (TyConstructor.Type "*", [ty1; ty2]) ->
      if level >= 1 then
        print_type_aux 0 ty1 ^^
        string " * " ^^
        print_type_aux 1 ty2
      else
        parens (print_type ty)
  | TyCon (Type "->", [ty1; ty2]) ->
      if level >= 2 then
        print_type_aux 1 ty1 ^^
        string " -> " ^^
        print_type ty2
      else
        parens (print_type ty)
  | TyCon (Type cons, tys) ->
    if level >= 2 then
      string cons ^^
      string " " ^^
      (
        if List.length tys > 0 then
          (parens (print_args level tys))
        else
         string ""
     )
    else parens (print_type ty)
  | TyMu (x, ty1) ->
      if level >= 3 then
        string "mu "  ^^
        print_tyvar x ^^
        dot ^^ space ^^
        print_type_aux 3 ty1
      else
        parens (print_type ty)

and print_type ty =
  print_type_aux 3 ty

(* -------------------------------------------------------------------------- *)

(* Terms. *)

let rec print_term_aux acc tab level t =
  assert (level >= 0);
  match t with
  | NTyVar (tyvar, t) ->
    print_term_aux (tyvar::acc) tab level t

  | _ ->
    print_newtyvar acc ^^
    (match t with
     | Var x ->
       string x

     | App (t1, t2) ->
       if level >= 1 then
         print_term_aux [] tab 1 t1 ^^
        space ^^
         print_term_aux [] tab 0 t2
       else
         parens (print_term tab t)

     | Abs (x, tyopt, t2) ->
       if level >= 2 then
         string "fun " ^^
         string x ^^
         (print_annot tyopt) ^^
         string " = " ^^
         print_term_aux [] tab 2 t2
       else
         parens (print_term tab t)

     | Let (x, tyopt, t1, t2) ->
      if level >= 2 then
        string "let " ^^
        string x ^^
        (print_annot tyopt) ^^
        string " = " ^^
        print_term tab t1 ^^
        string (" in \n"^(tab2string (tab+2))) ^^
        print_term_aux [] (tab+2) 2 t2
      else
        parens (print_term tab t)

     | Pair (t1, t2) ->
       parens (
         print_term tab t1 ^^
         comma ^^ space ^^
         print_term tab t2
       )

     | Proj (i, t2) ->
       (* like [App] *)
      if level >= 1 then
        string "proj" ^^
        OCaml.int i ^^
        space ^^
        print_term_aux [] tab 0 t2
      else
        parens (print_term tab t)

     | Lit (LInt i) ->
    string (string_of_int i)
     | Annot (t, ty) ->
    parens (print_term tab t ^^
            string " : " ^^
            print_type ty)

     | _ -> failwith "Can't happen"
    )

and print_term tab t =
  print_term_aux [] tab 2 t

and tab2string tab =
  String.init tab (fun i -> ' ')

and print_newtyvar tyvars =
  match tyvars with
  | [] -> PPrint.empty
  | x:: [] -> parens (string "new " ^^ print_tyvar x) ^^ string " "
  | _ -> parens (string "new " ^^ print_tyvars (List.rev tyvars)) ^^ string " "


let print_program p =

  List.fold_right
    (fun def doc ->
       match def with
       | TyDef (Type con, tyvars, Abbrev ty) ->
         string "type " ^^
         print_tyvars tyvars ^^
         string (" "^con^" = ") ^^
         print_type ty ^^
         string"\n\n" ^^
         doc

       | ValDef (x, t) ->
         string "let " ^^
         string x ^^
         string (" = \n"^(tab2string 2)) ^^
         print_term 2 t ^^
         string "\n\n" ^^
         doc
    )
    p
    PPrint.empty
