open Env

module L = struct
  type tyvar = ML.tyvar
  type typ = ML.typ

  module TyvarSet = Set.Make(struct type t = tyvar let compare = compare end)

  let free_vars (typedecl_env : (ML.tycon, 't TyConstructor.t) Env.t) ty =
    let rec free ty bound =
      match ty with
      | ML.TyVar v' ->
          if TyvarSet.mem v' bound
          then TyvarSet.empty
          else TyvarSet.singleton v'

      | ML.TyMu (v', ty') ->
          free ty' (TyvarSet.add v' bound)

      | ML.TyCon (cons, tys) ->
          let descr = Env.lookup cons typedecl_env in
          let decomp = TyConstructor.decomposable descr in
          assert (List.length tys = List.length decomp);
          List.fold_left2
            (fun bs d ty ->
              if d
              then TyvarSet.union (free ty bound) bs
              else bs)
            TyvarSet.empty
            decomp
            tys
    in
    free ty TyvarSet.empty

  let is_variable ty =
    match ty with
    | ML.TyVar v ->
        Some v
    | _ ->
        None

  let is_abbrev typedecl_env ty =
    match ty with
    | ML.TyCon (c, tys) ->
        let t = Env.lookup c typedecl_env in
        Some (t, tys)
    | _ ->
        None

end

open TyConstructor

module TyConstr = TyConstructor.Make(L)

(* The unifier will use the following type structure. *)
module S = struct

  module ME = MultiEq.Make(L)
  include ME

  module Deep = UnifierSig.MakeDeep(ME)

  let rec convert
    typedecl_env
    (* Environment linking type variable in ML and type variable in
       the constraint language. *)
    (varenv : (L.tyvar, 'var Deep.t) Env.t)
    (* Type to translate *)
    (tau  : L.typ)
    : 'var Deep.t
  =
    match tau with
      | ML.TyVar var ->
          Env.lookup var varenv
      | ML.TyCon (c, tys) ->
          let con = Env.lookup c typedecl_env in
          Structure (ME.init con (List.map (convert typedecl_env varenv) tys))
      | ML.TyMu (var, ty) ->
          failwith "TODO for later"

  let convert_abbrev typedecl_env ({cons; vars} : 'a Deep.t t) : 'a Deep.t =
    (*
      - on construit un varenv à partir de "vars" (on peut garder le code existante)
      - on appelle "convert" sur le "tau" extrait de l'expansion, définition ci-dessous
     *)

    (* First we need to extract the expansion of the abbreviation. *)
    let tau =
      match TyConstructor.expansion cons with
      | Some t -> t
      | None -> failwith "Not an abbreviation."
    in

    let varenv =
      let c = ref (-1) in
      List.fold_left
        (fun env v ->
           c:= !c + 1; Env.bind !c v env)
        Env.empty  vars in

    ( convert typedecl_env varenv tau )

  let resolve
      typedecl_env
      (meq : 'a structure) : ('a Deep.t * 'a Deep.t) list * 'a Deep.t structure
    =
    (* After an expansion, the new representative may create a
       decomposition with another view. *)
    (* For now, I suppose we have the invariant that all
       decompositions have been made during merging.  We also keep
       track of the variables on which a decomposition is done, to
       solve then again later. *)
    let to_unify : ('a Deep.t * 'a Deep.t) list ref = ref [] in

    let rec merge' (v : 'a Deep.t view) (s : 'a Deep.t structure) : 'a Deep.t structure =
      match s with
      | [] ->
         [v]
      | v' :: meq ->
        let cons_opt : 'a Deep.t t option = find_shared_constructor v v' in
        match cons_opt with

        (* No shared constructor *)
        | None ->
           v' :: merge' v meq

        (* We found a shared constructor [cons] *)
        | Some { cons ; vars } ->

           (* Decompose [v.repr] *)
           decompose v.repr.cons (fun t u -> to_unify := (t,u) :: !to_unify) v.repr.vars vars ;

           (* Finally, we merge [v] and [v']. The remaining of the
              multi-equation (meq) stays inchanged *)
           (merge_view v v') ::  meq
    in

    (* Find the view with the biggest rank of a multi-equation.
       Return this view along with the multi-equation minus this view. *)
    let find_bigger_rank (fstview : 'a Deep.t view) (meq : 'a Deep.t structure)
      : 'a Deep.t view * 'a Deep.t structure
      =
      List.fold_left
        (fun (max, acc) v ->
           if max.repr.cons <. v.repr.cons
           then (v, max::acc)
           else (max, v::acc))
        (fstview, []) meq
    in

    let rec resolve_ (meq : 'a Deep.t structure) : 'a Deep.t structure =
      match meq with
      (* Already solved *)
      | [] | [_] ->
        meq

      (* Unsolved *)
      | v :: vs ->

        let v, vs = find_bigger_rank v vs in

        (* Expansion *)
        (* 1. Compute the expansion type. *)
        let nmeq =
          match convert_abbrev typedecl_env v.repr with
          | Var _ -> failwith "Expansion étoilée : TODO"
          | Structure s -> s
        in

        (* 2. This term is the new representative of the view. *)
        let v = replace_repr v (List.hd nmeq).repr in

        (* 3. We now need to check if we can perform a decomposition. *)
        let nmeq = merge' v vs in
        resolve_ nmeq

    in

    let meq = resolve_ (map (fun t -> Deep.Var t) meq) in

    !to_unify, meq

end
