open TyConstructor

module Make (L : TyConstructor.LANG) = struct

  type 'a t  =
    { cons : L.typ TyConstructor.t ; vars : 'a list}

  type env = L.typ TyConstructor.env

  module VM = Map.Make (struct
    type t = L.typ TyConstructor.t
    let compare = compare
  end)

  (* An view :
     - repr : representative (terme with minimal rank in the multieq)
     - others : other type terms, already checked as equivalent by type inference,
       indexed here by their head constructor *)
  type 'a view = { repr : 'a t ; others : 'a t VM.t}

  (* A multi-equation. *)
  type 'a structure = ('a view) list

  (* -------------------------------------------------------------------------- *)

  let init c tys =
    let t = {
        cons = c;
        vars = tys;
      } in
    [
      { repr = t ; others = VM.empty }
    ]

  (* A multi-equation is in solved form if it is a singleton. *)
  let is_resolved t =
    List.length t = 1

  let from_singleton t =
    match t with
    | [x] ->
        x
    | _ ->
        failwith "get_repr : Not in solved form."

  (* The representative of a multi-equation exists if and only if the
     view is solved and is the representative of the unique view. *)
  let get_repr t =
    let x = from_singleton t in
    x.repr

  let set_repr meq nrepr =
    {meq with repr = nrepr}

  (* [dispatch cons vars] place the small term {cons;vars} in a
     singleton view. *)
  let dispatch cons vars : 'a structure =
    [{repr = {cons ; vars} ; others = VM.empty}]

  let map_view f {repr; others} =
    let repr =
      { cons = repr.cons ; vars = List.map f repr.vars} in
    let others =
      VM.map (fun {cons; vars} -> {cons; vars = List.map f vars}) others in
    {repr ; others}

  let map (f : 'a -> 'b) (t : 'a structure) : 'b structure =
    List.map (map_view f) t

  let fold f (t : 'a structure) accu =
    (* t should be a singleton (solved form). *)
    let meq = from_singleton t in
    let r = meq.repr in
    List.fold_right f r.vars accu

  let iter f (t : 'a structure) =
    (* t should be a singleton (solved form). *)
    let meq = from_singleton t in
    let r = meq.repr in
    List.iter f r.vars

  let incompatible_struct (t : 'a structure) (u : 'a structure)  =
      match t, u with
      | v1 :: _ , v2 :: _ ->
          incompatible v1.repr.cons v2.repr.cons
      | _, _ ->
          false

  (* Two views are merged when a decomposition between two of their
       elements has been made. *)
  let rec merge_view v1 v2 =
    if v1.repr.cons <=. v2.repr.cons
    then
      { repr =
          (* Representative : the one with the smaller rank is kept *)
          v1.repr;
        others =
          (* To compute the others equivalent types, we perform an
             union without forgetting the representative of the second
             view. If a constructor is present in both view, one is
             kept arbitrarily. *)
          let v2oth =
            if v1.repr.cons == v2.repr.cons
            then v2.others
            else VM.add v2.repr.cons v2.repr v2.others
          in
          VM.union (fun _ ty1 _ -> Some ty1) v1.others v2oth }
    else merge_view v2 v1

  (* Before merging two views, we need to find the shared
     constructors. *)
  let find_shared_constructor v v' : 'a t option =
    if v.repr.cons >. v'.repr.cons
    then VM.find_opt v.repr.cons v'.others
    else if v.repr.cons <. v'.repr.cons
    then VM.find_opt v'.repr.cons v.others
    else if equal v.repr.cons v'.repr.cons
    then Some v'.repr
    else None

  let merge_decompose f (t : 'a structure) (u : 'a structure) =

    let rec merge_ flag acc v meq = match meq with
      | [] -> flag, v, acc
      | v' :: meq ->
        let cons_opt = find_shared_constructor v v' in
        match cons_opt with
        | None ->
          merge_ flag (v'::acc) v meq
        | Some ty ->
          decompose v.repr.cons f v.repr.vars ty.vars ;
          let nv = merge_view v v' in
          merge_ true acc nv meq
    in

    let not_decomposed, merged =
    List.fold_left
      (fun (acc, u) vt ->
         let flag, v, u = merge_ false acc vt u in
         if flag then
           (acc, v::u)
         else
           (v::acc, u)
      )
      ([], u) t
    in
    not_decomposed @ merged

  exception Iter2

  (* Merge *)
  let merge f (t : 'a structure) (u : 'a structure) : 'a structure  =
    if incompatible_struct t u
    then raise Iter2
    else merge_decompose f t u

  let replace_repr {repr; others} new_repr =
    {others = VM.add repr.cons repr others ; repr = new_repr}

end
