open Env

type id = Type of string

type 't t = {
    id : id ;
    arity : int ;

    (* Each element of the list indicates if the corresponding parameter appears
       in the expansion of the constructor. *)
    decomposable : bool list ;
    rank : int ;
    expansion : 't option ;
    head : id option ;
    star : bool ;
  }

(* Getters *)
let arity c =  c.arity

let name c = c.id

let decomposable c = c.decomposable

let rank c = c.rank

let head c = c.head

let expansion c = c.expansion

(* -------------------------------------------------------------------------- *)

type 't env = (id, 't t) Env.t

let empty =
  Env.empty

let cardinal typedecl_env =
  Env.cardinal typedecl_env

let lookup x typedecl_env =
  Env.lookup x typedecl_env

let bind x v typedecl_env =
  Env.bind x v typedecl_env

(* -------------------------------------------------------------------------- *)

(* If [incompatible t u] is true then the type constructor [t] and [u]
     are not compatible (it is not possible to reach one by expanding
     the other). If it is [false], then t and u may be compatible. *)
let incompatible t u =
  match t.head, u.head with
  | Some th, Some uh when th = uh ->
      false
  | None, None ->
      false
  | _, _ ->
      true


(* This function will need to be changed if it become possible to
     define multiple type with the same name. *)
let equal t u =
  t.id = u.id

let compare t u =
  compare t.id u.id

let (>.) t u = t.rank > u.rank
let (<.) t u = t.rank < u.rank
let (>=.) t u = t.rank >= u.rank
let (<=.) t u = t.rank <= u.rank
let (=.) t u = t.rank = u.rank

let rec decompose t f us ts =
  let d = t.decomposable in
  assert (t.arity = List.length d && t.arity = List.length ts && t.arity = List.length us);
  let rec aux d ts us  =
    match d, ts, us with
    | [], [], [] -> ()
    | true  :: d, t :: ts, u :: us -> f t u ; aux d ts us
    | false :: d, _ :: ts, _ :: us -> aux d ts us
    | [], _, _
    | _, [], _
    | _, _, [] -> assert false

  in
  aux d us ts

let build_type_cons id arity =
  {id ;
   arity ;
   rank=0 ;
   expansion = None ;
   decomposable = List.init arity (fun _ -> true) ;
   head = Some id ;
   star = false ;}


let arrow = build_type_cons (Type "->") 2

let product = build_type_cons (Type "*") 2

(* -------------------------------------------------------------------------- *)

module type LANG = sig
  type tyvar
  type typ

  module TyvarSet : Set.S with type elt = tyvar
  val free_vars : typ env -> typ -> TyvarSet.t

  val is_variable : typ -> tyvar option
  val is_abbrev : typ env -> typ -> (typ t * typ list) option
end

(* -------------------------------------------------------------------------- *)

module Make (L : LANG) = struct

  let build_abbrev_cons typedecl_env cons vars def =

    let used_vars = L.free_vars typedecl_env def in
    assert (L.TyvarSet.subset used_vars (L.TyvarSet.of_list vars));

    let decomposable =
      List.map
        (fun v -> L.TyvarSet.mem v used_vars)
        vars
    in

    let rec is_star ty =
      match L.is_variable ty with
      | Some _ ->
          true
      | None ->
         match L.is_abbrev typedecl_env ty with
         | None ->
             false
         | Some (c, tys) ->
             if not c.star then
               false
             else
               List.fold_left2
                 (fun b d ty ->
                   if d then b && is_star ty
                   else b)
                 true
                 c.decomposable
                 tys
    in

    let star = is_star def in

    let rank, head  =
      match L.is_abbrev typedecl_env def with
      | None ->
          1, None
      | Some (c, tys) ->
          c.rank + 1, c.head
    in

    {id = cons ;
     arity = List.length vars ;
     expansion = Some def;
     rank ;
     decomposable ;
     head ;
     star}

end
